/*
	Genetic Algorithm Simulation for TalesWeaver
	Simulation Parameter Module
*/
////////////////////////////////////////////////////////////////////////

#include "../common/stdafx.h"
#include "GAparam.h"

// ϐ̊S
void CGAparam::initialize()
{
	// LN^[
	m_chara = 0;
	m_style = 0;
	m_level = 100;

	// ڕWXe[^X
	for (int i=0; i<NUM_STAT; i++) {
		m_stat_min[i]  = 0;
		m_stat_max[i]  = MAX_STATUS;
		m_stat_wei[i]  = 1;
		m_init_stat[i] = 0;
	}
	m_stat_wei_hp = 0.01;
	m_stat_wei_mp = 0.01;

	// Xe[^X
	m_init_use   = false;
	m_init_level = 0;
	m_init_point = 0;

	// LXienEXe[^X
	m_ignored_xien = 0;
	m_ignored_stat = 0;

	// ʃIvV
	m_opt_output_fix   = true;
	m_opt_random_rate  = 50;
	m_opt_penalty_xien = 0.5;
	m_opt_penalty_stat = 0.1;

	// GAIvV
	m_ga_try_time    = 1;
	m_ga_generation  = 10000;
	m_ga_unit_elite  = 1;
	m_ga_unit_twice  = 100;
	m_ga_unit_unif   = 100;
	m_ga_unit_swap   = 100;
	m_ga_rank_param  = 50;
	m_ga_mutate_rate = 0.1;

	need_point_new();
	m_proc_time = 0;
}

// p[^`FbN
bool CGAparam::validate(char *errmsg)
{
	// LN^[
	if (IS_OUT_OF_RANGE(m_chara, 0, NUM_CHARA-1)) {
		sprintf(errmsg, "LN^[^Cv(%d)sł", m_chara);
		return false;
	}
	if (IS_OUT_OF_RANGE(m_style, m_chara * NUM_STYLE, m_chara * NUM_STYLE + 2)) {
		sprintf(errmsg, "LN^[X^C(%d)sł", m_style);
		return false;
	}
	ASSERT_BY_RANGE_B(errmsg, m_level, "LN^[x", 2, MAX_LEVEL);
	m_sim_init_xien = style_xien(m_style);
	for (int i=0; i<NUM_XIEN; i++) for (int j=0; j<NUM_SKILL; j++) m_sim_skill[i][j] = skill_level_find(m_chara, i, j);

	// ڕWXe[^X
	for (int i=0; i<NUM_STAT; i++) {
		char str_min[16];
		char str_max[16];
		char str_wei[16];
		sprintf(str_min, "%sڕWl", stat_name(i));
		sprintf(str_max, "%sől", stat_name(i));
		sprintf(str_wei, "%sdvx", stat_name(i));
		ASSERT_BY_RANGE_B(errmsg, m_stat_min[i], str_min, 0, MAX_STATUS);
		ASSERT_BY_RANGE_B(errmsg, m_stat_max[i], str_max, 0, MAX_STATUS);
		ASSERT_BY_RANGE_G(errmsg, m_stat_wei[i], str_wei, 0);
		m_stat_min_100[i] = m_stat_min[i] * 100;
		m_stat_max_100[i] = m_stat_max[i] * 100;
	}
	initialize_stat_minmax(m_stat_min, m_stat_max);

	// Xe[^X
	if (m_init_use) {
		ASSERT_BY_RANGE_B(errmsg, m_init_level, "ԃLN^[x", 1, m_level - 1);
		ASSERT_BY_RANGE_G(errmsg, m_init_point, "ԃXe[^X|Cg", 0);
		for (int i=0; i<NUM_STAT; i++) {
			char str_init[16];
			sprintf(str_init, "%sl", stat_name(i));
			ASSERT_BY_RANGE_B(errmsg, m_init_stat[i], str_init, 0, MAX_STATUS);
		}
		m_sim_level_start = m_init_level + 1;
		m_sim_level_final = m_level;
		m_sim_point       = m_init_point;
		for (int i=0; i<NUM_STAT; i++) m_sim_stat_init_100[i] = m_init_stat[i] * 100;
		for (int i=0; i<NUM_STAT; i++) m_sim_stat_req[i]      = style_req(m_style, i);
	} else {
		m_sim_level_start = 2;
		m_sim_level_final = m_level;
		m_sim_point       = 0;
		for (int i=0; i<NUM_STAT; i++) m_sim_stat_init_100[i] = style_stat(m_style, i) * 100;
		for (int i=0; i<NUM_STAT; i++) m_sim_stat_req[i]      = style_req(m_style, i);
	}

	// LXienEXe[^X
	m_available_xien = 0;
	m_available_stat = 0;
	for (int i=0; i<NUM_XIEN; i++) if (m_sim_skill[i][0]) m_available_xien |= 1 << i;
	for (int i=0; i<NUM_STAT; i++)                        m_available_stat |= 1 << i;
	if (m_ignored_xien) m_available_xien &= ~m_ignored_xien;
	if (m_ignored_stat) m_available_stat &= ~m_ignored_stat;

	// ʃIvV
	ASSERT_BY_RANGE_B(errmsg, m_opt_random_rate, "_㏸", 0, 100);
	m_opt_penalty_xien_I = (int)(m_opt_penalty_xien * 100.0);
	m_opt_penalty_stat_I = (int)(m_opt_penalty_stat * 100.0);

	// GAIvV
	ASSERT_BY_RANGE_G(errmsg, m_ga_try_time,    "GAs",           1);
	ASSERT_BY_RANGE_G(errmsg, m_ga_generation,  "GA㐔",             2);
	ASSERT_BY_RANGE_B(errmsg, m_ga_unit_elite,  "G[gI̐",   0, 1);
	ASSERT_BY_RANGE_G(errmsg, m_ga_unit_twice,  "_̐",       0);
	ASSERT_BY_RANGE_G(errmsg, m_ga_unit_unif,   "l̐",       0);
	ASSERT_BY_RANGE_G(errmsg, m_ga_unit_swap,   "zu]̐",       0);
	ASSERT_BY_RANGE_B(errmsg, m_ga_rank_param,  "LOIx",   0, 100);
	m_ga_unit_total = m_ga_unit_elite + m_ga_unit_twice + m_ga_unit_unif + m_ga_unit_swap;
	if (IS_OUT_OF_RANGE(m_ga_try_time, 0, m_ga_unit_total)) {
		sprintf(errmsg, "%s(%d)͌̘̐aȉłȂ΂Ȃ܂", "GAs", m_ga_try_time);\
		return false;
	}
	m_ga_rank_tsize = 0;
	for (int i=0; i<m_ga_unit_total; i++) {
		int n = m_ga_unit_total - i * m_ga_rank_param / 100;
		if (n < 1) n = 1;
		m_ga_rank_tsize += n;
	}
	if (m_ga_rank_table) delete [] m_ga_rank_table;
	m_ga_rank_table = new int[m_ga_rank_tsize];
	for (int i=0, k=0; i<m_ga_unit_total; i++) {
		int n = m_ga_unit_total - i * m_ga_rank_param / 100;
		if (n < 1) n = 1;
		for (int j=0; j<n; j++, k++) m_ga_rank_table[k] = i;
	}
	int lvup_num = m_sim_level_final - m_sim_level_start + 1;
	double f = 0.0;
	double fsum = 0.0;
	for (int r=0; r<=lvup_num; r++) {
		if (r == 0) f = pow(1.0 - m_ga_mutate_rate / 100.0, (double)lvup_num);
		else        f *= m_ga_mutate_rate / (100.0 - m_ga_mutate_rate) * (lvup_num - r + 1) / r;
		fsum += f;
		m_ga_mutate_table[r] = (int)(fsum * 16777216.0) + 1;
	}

	return true;
}

void CGAparam::initialize_stat_minmax(int *stat_min, int *stat_max)
{
	for (int i=0; i<NUM_STAT; i++) {
		m_stat_min_100[i] = stat_min[i] * 100;
		m_stat_max_100[i] = stat_max[i] * 100;
	}
}

// Kv|Cge[u쐬
void CGAparam::need_point_new()
{
	if (m_need_point_100) return;

	m_need_point_100 = new int [(MAX_LEVEL + 1) * 65536];
	for (int lv=0; lv<=MAX_LEVEL; lv++) {
		for (int s=0; s<65536; s++) {
			need_point_100(lv, s) = need_point(lv, s / 100 > 255 ? 255 : s / 100);
		}
	}
}

// Kv|Cge[u폜
void CGAparam::need_point_delete()
{
	if (!m_need_point_100) return;

	delete [] m_need_point_100;
	m_need_point_100 = NULL;
}
