/*
	Genetic Algorithm Simulation for TalesWeaver
	Genom Unit Module
*/
////////////////////////////////////////////////////////////////////////

#include "../common/stdafx.h"
#include "GAunit.h"

CGAparam	*CGAunit::m_param = NULL;
ETYPE		CGAunit::m_backup_genom[NUM_ELEM];
int			CGAunit::m_backup_stat[NUM_ELEM][NUM_STAT];
int			CGAunit::m_backup_hpmp[NUM_ELEM][2];
int			CGAunit::m_backup_point[NUM_ELEM];

////////////////////////////////////////////////////////////////////////

// RXgN^
CGAunit::CGAunit()
{
	memset(m_genom, 0, sizeof(m_genom));
	m_evaled = false;
}

// Rs[RXgN^
CGAunit::CGAunit(CGAunit &src)
{
	*this = src;
}

// 
CGAunit &CGAunit::operator=(CGAunit &src)
{
	memcpy(m_genom, src.m_genom, sizeof(m_genom));
	memcpy(m_stat,  src.m_stat,  sizeof(m_stat));
	memcpy(m_hpmp,  src.m_hpmp,  sizeof(m_hpmp));
	m_point  = src.m_point;
	m_score  = src.m_score;
	m_evaled = src.m_evaled;
	return *this;
}

// r
bool operator==(CGAunit &src1, CGAunit &src2)
{
	int i;
	ETYPE *seq1 = src1.m_genom;
	ETYPE *seq2 = src2.m_genom;
	for (i=0; i<NUM_ELEM; i++) if (*(seq1++) != *(seq2++)) return false;
	return true;
}

////////////////////////////////////////////////////////////////////////

// `vf(S_)
ETYPE CGAunit::element()
{
	ETYPE elem_xien = element_xien();
	ETYPE elem_stat = element_stat();
	return elem_xien | elem_stat;
}

// `vf(ˑRψ)
ETYPE CGAunit::element_mutate(ETYPE elem_base)
{
	if (rand_bool()) return element_xien() | (elem_base & ELEM_MASK_STAT);
	else             return element_stat() | (elem_base & ELEM_MASK_XIEN);
}

// `vf(XIEN)
ETYPE CGAunit::element_xien()
{
	while (1) {
		ETYPE idx = rand_range1(NUM_XIEN);
		if ((1 << idx) & D_AVAILABLE_XIEN) return conv_idx2elem_xien(idx);
	}
}

// `vf(STAT)
ETYPE CGAunit::element_stat()
{
	while (1) {
		ETYPE idx = rand_range1(NUM_STAT);
		if ((1 << idx) & D_AVAILABLE_STAT) return conv_idx2elem_stat(idx);
	}
}

////////////////////////////////////////////////////////////////////////

// `q
void CGAunit::genom_init()
{
	memset(m_genom, 0, sizeof(m_genom));
	for (int i=D_LEVEL_START; i<=D_LEVEL_FINAL; i++) m_genom[i] = element();
	m_evaled = false;
}

// _
void CGAunit::genom_twice(CGAunit &src1, CGAunit &src2)
{
	int pos1;
	int pos2;
	do {
		pos1 = rand_range2(D_LEVEL_START + 1, D_LEVEL_FINAL);
		pos2 = rand_range2(D_LEVEL_START + 1, D_LEVEL_FINAL);
	} while (pos1 >= pos2);

	for (int i=D_LEVEL_START; i<pos1;           i++) m_genom[i] = src1.m_genom[i];
	for (int i=pos1;          i<pos2;           i++) m_genom[i] = src2.m_genom[i];
	for (int i=pos2;          i<=D_LEVEL_FINAL; i++) m_genom[i] = src1.m_genom[i];

	genom_mutate();
	m_evaled = false;
}

// l
void CGAunit::genom_unif(CGAunit &src1, CGAunit &src2)
{
	for (int i=D_LEVEL_START; i<=D_LEVEL_FINAL; i++) m_genom[i] = rand_bool() ? src1.m_genom[i] : src2.m_genom[i];

	genom_mutate();
	m_evaled = false;
}

// zu]
void CGAunit::genom_swap(CGAunit &src)
{
	*this = src;

	int pos1;
	int pos2;
	do {
		if (rand_bool()) {
			pos1 = rand_range2(D_LEVEL_START, D_LEVEL_FINAL - 1);
			pos2 = pos1 + 1;
		} else {
			pos1 = rand_range2(D_LEVEL_START, D_LEVEL_FINAL);
			pos2 = rand_range2(D_LEVEL_START, D_LEVEL_FINAL);
		}
	} while (pos1 == pos2);
	genom_modify_swap_element(pos1, pos2);

	genom_mutate();
	m_evaled = false;
}

// gˑRψق
void CGAunit::genom_mutate()
{
	int mutate_rand = rand_b24();
	int mutate_count;
	for (mutate_count=0; mutate_count<NUM_ELEM; mutate_count++) {
		if (mutate_rand < D_GA_MUTATE_TABLE[mutate_count]) break;
	}

	if (mutate_count) {
		for (int i=0; i<mutate_count; i++) {
			int pos = rand_range2(D_LEVEL_START, D_LEVEL_FINAL);
			m_genom[pos] = element_mutate(m_genom[pos]);
		}
		m_evaled = false;
	}
}

// `q̗vfւ(P)
void CGAunit::genom_modify_swap_element(int pos1, int pos2)
{
	ETYPE tmp     = m_genom[pos1];
	m_genom[pos1] = m_genom[pos2];
	m_genom[pos2] = tmp;
	m_evaled = false;
}

// `q̗vfւ(͈)
void CGAunit::genom_modify_swap_band(int pos11, int pos12, int pos21, int pos22)
{
	if (!(pos11 < pos12 && pos12 < pos21 && pos21 < pos22)) return;
	
	ETYPE tmp[NUM_ELEM];
	int pos = pos11;
	for (int i=pos21;   i<=pos22; i++) tmp[pos++] = m_genom[i];
	for (int i=pos12+1; i< pos21; i++) tmp[pos++] = m_genom[i];
	for (int i=pos11;   i<=pos12; i++) tmp[pos++] = m_genom[i];
	for (int i=pos11;   i<=pos22; i++) m_genom[i] = tmp[i];
	m_evaled = false;
}

////////////////////////////////////////////////////////////////////////

// `qAz񂩂NX֕ϊ
void CGAunit::genomband_import(int &vec_size, int *vec_lv1, int *vec_lv2, ETYPE *vec_elem)
{
	for (int i=0; i<vec_size; i++) {
		for (int j=vec_lv1[i]; j<=vec_lv2[i]; j++) m_genom[j] = vec_elem[i];
	}
	m_evaled = false;
}

// `qANXz֕ϊ
void CGAunit::genomband_export(int &vec_size, int *vec_lv1, int *vec_lv2, ETYPE *vec_elem)
{
	vec_size = 0;

	int last_level = D_LEVEL_START;
	ETYPE last_elem = m_genom[last_level];
	for (int lv=D_LEVEL_START+1; lv<=D_LEVEL_FINAL; lv++) {
		if (m_genom[lv] != last_elem) {
			vec_lv1 [vec_size] = last_level;
			vec_lv2 [vec_size] = lv - 1;
			vec_elem[vec_size] = last_elem;
			vec_size++;
			last_level = lv;
			last_elem  = m_genom[lv];
		}
	}
	vec_lv1 [vec_size] = last_level;
	vec_lv2 [vec_size] = D_LEVEL_FINAL;
	vec_elem[vec_size] = last_elem;
	vec_size++;
}

////////////////////////////////////////////////////////////////////////

// Kx]
int CGAunit::eval(bool do_force, int *hmsw)
{
	// ɕ]Ă炻̒lԂ
	if (m_evaled && !do_force) return m_score;

	// Xe[^X
	int lv = D_LEVEL_START;
	eval_init();

	// obNAbv烍[h
	if (!do_force) eval_backup_load(lv);

	// LvƂɃ[vďԂvZ
	int count_xien[NUM_XIEN];
	int count_stat[NUM_STAT];
	int *p_count_xien = NULL;
	int *p_count_stat = NULL;
	if (hmsw) {
		memset(count_xien, 0, sizeof(count_xien));
		memset(count_stat, 0, sizeof(count_stat));
		p_count_xien = count_xien;
		p_count_stat = count_stat;
	}
	for (; lv<=D_LEVEL_FINAL; lv++) {
		// Xe[^X|Cgl
		m_point += m_param->lvup_point(lv);
		// p[^l
		eval_status_auto(lv, p_count_xien);
		// 蓮p[^l
		eval_status_manual(lv, p_count_stat);
		// obNAbvL^
		if (do_force) eval_backup_save(lv);
	}

	// I[o[t[
	eval_check_overflow();

	// HP/MP/SP/WTmzvZ
	if (hmsw) calc_hmsw_distribution(hmsw, count_xien, count_stat);

	// ]
	m_score = 0;
	eval_score_point();
	eval_score_status();
	eval_score_hmsw();
	eval_score_penarty();
	eval_score_check_xien();
	m_evaled = true;
	return m_score;
}

// Xe[^X
void CGAunit::eval_init()
{
	for (int i=0; i<NUM_STAT; i++) {
		m_stat[i]   = D_INIT_STAT[i];
	}
	m_hpmp[INDEX_HP] = 0;
	m_hpmp[INDEX_MP] = 0;
	m_point = D_INIT_POINT;
}

// Xe[^XZ[u
void CGAunit::eval_backup_save(int &lv)
{
	m_backup_genom[lv] = m_genom[lv];
	for (int i=0; i<NUM_STAT; i++) m_backup_stat[lv][i] = m_stat[i];
	for (int i=0; i<2;        i++) m_backup_hpmp[lv][i] = m_hpmp[i];
	m_backup_point[lv] = m_point;
}

// `qʂĂ͈͂ł̃Xe[^X[h
void CGAunit::eval_backup_load(int &lv)
{
	for (lv=D_LEVEL_START; lv<=D_LEVEL_FINAL; lv++) if (m_genom[lv] != m_backup_genom[lv]) break;
	if (lv == D_LEVEL_START) return;
	for (int i=0; i<NUM_STAT; i++) m_stat[i] = m_backup_stat[lv-1][i];
	for (int i=0; i<2;        i++) m_hpmp[i] = m_backup_hpmp[lv-1][i];
	m_point = m_backup_point[lv-1];
}

// Xe[^XUP
void CGAunit::eval_status_auto(int &lv, int *count_xien)
{
	int xien = conv_elem2idx_xien(m_genom[lv]);
	m_stat[m_param->xien_up_fix(xien)] += 100;
	m_stat[m_param->xien_up_rnd(xien)] += D_OPT_RANDOM_RATE;
	m_hpmp[INDEX_HP] += m_param->m_hmsw_sum_auto[INDEX_HP][xien];
	m_hpmp[INDEX_MP] += m_param->m_hmsw_sum_auto[INDEX_MP][xien];
	if (count_xien) count_xien[xien]++;
}

// 蓮Xe[^XUP
void CGAunit::eval_status_manual(int &lv, int *count_stat)
{
	int stat = conv_elem2idx_stat(m_genom[lv]);
	while (1) {
		int point_need = eval_need_point(lv, stat);
		if (m_point < point_need) break;
		m_point -= point_need;
		m_stat[stat] += 100;
		m_hpmp[INDEX_HP] += m_param->m_hmsw_sum_manual[INDEX_HP][stat];
		m_hpmp[INDEX_MP] += m_param->m_hmsw_sum_manual[INDEX_MP][stat];
		if (count_stat) count_stat[stat]++;
	}
}

// Xe[^X̃I[o[t[`FbN
void CGAunit::eval_check_overflow()
{
	for (int i=0; i<NUM_STAT; i++) {
		// Xe[^Xől255
		if (m_stat[i] > MAX_STATUS * 100) m_stat[i] = MAX_STATUS * 100;
	}
}

// KxZo(c|Cg)
// c|Cg1œKx10(Xe[^X0.1xdvx1)Z
// Xe[^XAbvɂlXe[^XZ|Cg1œKx10Z
void CGAunit::eval_score_point()
{
	m_score += m_point * 10;
}

// KxZo(Xe[^X)
// Xe[^X1xdvx1œKx100Z
void CGAunit::eval_score_status()
{
	for (int i=0; i<NUM_STAT; i++) if (D_STAT_WEI[i]) {
		if      (m_stat[i] < D_STAT_MIN_100[i]) m_score += (m_stat[i]         - D_STAT_MIN_100[i]) * 10;
		else if (m_stat[i] < D_STAT_MAX_100[i]) m_score += (m_stat[i]         - D_STAT_MIN_100[i]) * D_STAT_WEI[i];
		else                                    m_score += (D_STAT_MAX_100[i] - D_STAT_MIN_100[i]) * D_STAT_WEI[i];
	}
}

// KxZo(őHP/MP)
// HP/MP1xdvx1œKx100Z
void CGAunit::eval_score_hmsw()
{
	m_score += (int)((m_hpmp[INDEX_HP] * 50) * D_STAT_WEI_HP);
	m_score += (int)((m_hpmp[INDEX_MP] * 50) * D_STAT_WEI_MP);
}

// KxZo(؂ւyieB)
void CGAunit::eval_score_penarty()
{
	for (int lv=D_LEVEL_START+1; lv<=D_LEVEL_FINAL; lv++) {
		if ((m_genom[lv] & ELEM_MASK_XIEN) != (m_genom[lv-1] & ELEM_MASK_XIEN)) {
			m_score -= D_OPT_PENALTY_XIEN;
		}
		if ((m_genom[lv] & ELEM_MASK_STAT) != (m_genom[lv-1] & ELEM_MASK_STAT)) {
			m_score -= D_OPT_PENALTY_STAT;
		}
	}
}

// KxZo(Is\ȐsXien)
void CGAunit::eval_score_check_xien()
{
	int last_xien  = D_SIM_INIT_XIEN;
	int last_skill = 0;

	for (int lv=D_LEVEL_START; lv<=D_LEVEL_FINAL; lv++) {
		int xien = conv_elem2idx_xien(m_genom[lv]);
		if (last_xien != xien) {
			if (xien < last_xien) last_skill++;
			if (last_skill == NUM_SKILL) return;
			last_xien = xien;
			if (lv <= D_SIM_SKILL[last_xien][last_skill]) m_score -= 10000;
		}
	}
}

// HP/MP/SP/WTmzvZ({)
void CGAunit::calc_hmsw_distribution(int *hmsw, int *count_xien, int *count_stat)
{
	int hmsw_limit[4] = {20000, 5000, 9999, 9999};
	memset(hmsw, 0, sizeof(int) * NUM_HMSW * 5);

	for (int i=0; i<NUM_HMSW; i++) {
		// Up
		int v_min[NUM_XIEN];
		int v_max[NUM_XIEN];
		for (int j=0; j<NUM_XIEN; j++) {
			v_min[j] = m_param->xien_hmsw_min(j, i);
			v_max[j] = m_param->xien_hmsw_max(j, i);
			hmsw[i * 5 + INDEX_HMSW_MIN] += v_min[j] * count_xien[j];
			hmsw[i * 5 + INDEX_HMSW_MAX] += v_max[j] * count_xien[j];
		}
		calc_hmsw_distribution_core(NUM_XIEN, v_min, v_max, count_xien, 99.99, hmsw[i * 5 + INDEX_HMSW_L99], hmsw[i * 5 + INDEX_HMSW_H99]);
		// 蓮Up
		for (int j=0; j<NUM_STAT; j++) {
			hmsw[i * 5 + INDEX_HMSW_MIN] += m_param->stat_hmsw(j, i) * count_stat[j];
			hmsw[i * 5 + INDEX_HMSW_MAX] += m_param->stat_hmsw(j, i) * count_stat[j];
			hmsw[i * 5 + INDEX_HMSW_L99] += m_param->stat_hmsw(j, i) * count_stat[j];
			hmsw[i * 5 + INDEX_HMSW_H99] += m_param->stat_hmsw(j, i) * count_stat[j];
		}
		// l
		if (!D_INIT_USE) {
			hmsw[i * 5 + INDEX_HMSW_MIN] += m_param->style_hmsw(m_param->m_style, i);
			hmsw[i * 5 + INDEX_HMSW_MAX] += m_param->style_hmsw(m_param->m_style, i);
			hmsw[i * 5 + INDEX_HMSW_L99] += m_param->style_hmsw(m_param->m_style, i);
			hmsw[i * 5 + INDEX_HMSW_H99] += m_param->style_hmsw(m_param->m_style, i);
		}
		// ϒl
		hmsw[i * 5 + INDEX_HMSW_AVG] = (hmsw[i * 5 + INDEX_HMSW_MIN] + hmsw[i * 5 + INDEX_HMSW_MAX]) / 2;
		// l
		if (hmsw[i * 5 + INDEX_HMSW_MIN] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_MIN] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_MAX] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_MAX] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_L99] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_L99] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_H99] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_H99] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_AVG] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_AVG] = hmsw_limit[i];
	}
}

// HP/MP/SP/WTmzvZ(RA)
void CGAunit::calc_hmsw_distribution_core(int v_size, int *v_min, int *v_max, int *v_num, double dist_range, int &out_min, int &out_max)
{
	int i, j;
	int loop1, loop2;

	// vZƊme[u쐬
	int limit = 0;
	for (i=0; i<v_size; i++) limit += v_max[i] * v_num[i];
	double *dist = new double[limit + 1];
	dist[0] = 1.0;
	for (i=1; i<=limit; i++) dist[i] = 0.0;
	int dist_min = 0;
	int dist_max = 0;

	// zvZ
	for (loop1=0; loop1<v_size; loop1++) {
		int e_min = v_min[loop1];
		int e_max = v_max[loop1];
		int e_num = v_num[loop1];
		for (loop2=0; loop2<e_num; loop2++) {
			int dist_min_next = dist_min + e_min;
			int dist_max_next = dist_max + e_max;
			for (i=dist_max_next; i>=dist_min_next; i--) {
				double tmp = 0.0;
				int stt = i - e_max;
				int end = i - e_min;
				if (stt < dist_min) stt = dist_min;
				if (end > dist_max) end = dist_max;
				for (j=stt; j<=end; j++) tmp += dist[j];
				dist[i] = tmp / 4.0;
			}
			dist_min = dist_min_next;
			dist_max = dist_max_next;
		}
	}

	// ݌vmƐK
	for (i=dist_min+1; i<=dist_max; i++) dist[i] += dist[i-1];
	for (i=dist_min  ; i<=dist_max; i++) dist[i] /= dist[dist_max];

	// z؂o
	double dist_range_min = (1.0 - dist_range / 100.0) / 2.0;
	double dist_range_max = (1.0 + dist_range / 100.0) / 2.0;
	for (out_min=dist_min; out_min<=dist_max; out_min++) if (dist[out_min] > dist_range_min) break;
	for (out_max=dist_max; out_max>=dist_min; out_max--) if (dist[out_max] < dist_range_max) break;
	if (out_min > dist_min) out_max--;
	if (out_max < dist_max) out_max++;

	delete [] dist;
}
