/*
	Genetic Algorithm Simulation for TalesWeaver
	Copyright (c) jeak@Anomarad 2007-2010
*/

////////////////////////////////////////////////////////////////////////

#include "../common/stdafx.h"
#include "../core/GAcore.h"
#include "../core/GAimex.h"

// 錾
void read_config(CGAparam &param, CGAunit &unit, char *config);
void set_param(CGAparam &param, CGAunit &unit, char *pname, char *pval);
void import_log(CGAparam &param, CGAunit &unit, char *logname, int mode);
void debug_print(CGAparam &param);

bool is_imported;

////////////////////////////////////////////////////////////////////////
// C[`

// C֐
int main(int argc, char **argv)
{
	printf("Genetic Algorithm Simulation for TalesWeaver\n");
	printf("Version %s\n", GASIM_VERSION);

	// ϐ̊S
	CGAparam param;
	CGAcore  core;
	CGAunit  unit_in;
	CGAunit  unit_out;
	core.initialize(param);

	// p[^ݒ
	is_imported = false;

	bool using_c_opt = false;
	for (int i=1; i<argc; i+=2) if (!strcmp(argv[i], "-c")) {
		using_c_opt = true;
		break;
	}
	if (!using_c_opt) read_config(param, unit_in, "GAsimC.cfg");

	for (int i=1; i<argc; i+=2) {
		char *pname = argv[i];
		char *pval  = argv[i+1];
		if (!strcmp(pname, "-c")) read_config(param, unit_in, pval);
		else                      set_param(param, unit_in, pname, pval);
	}

	// p[^`FbN
	char errmsg[256];
	if (!core.validate(errmsg)) {
		printf("----------------------------------------------------------------------\n");
		printf("%s\n", errmsg);
		return 0;
	}

//	debug_print(param);

	// GAs
	core.genetic_entrypoint(unit_in, unit_out, is_imported);

	// ʕ\
	GAimex *imex = (GAimex *)new GAimex_gasim;
	char log[65536];
	imex->export_log(param, unit_out, log);
	puts("---------------------------------------------\n");
	puts(log);
	delete imex;

	return 0;
}

// ݒt@C[h
void read_config(CGAparam &param, CGAunit &unit, char *config)
{
	// O
	FILE *fp = fopen(config, "r");
	if (!fp) {
		printf("error : ݒt@CJ܂ : %s\n", config);
		exit(-1);
	}
	// ڃ[h
	char buf[1024];
	char pname[1024];
	char pval[1024];
	while (fgets(buf, 1024, fp)) {
		if (buf[0] == '#' || buf[0] == '\r' || buf[0] == '\n') continue;
		sscanf(buf, "%s %s", pname, pval);
		set_param(param, unit, pname, pval);
	}
	// ㏈
	fclose(fp);
}

// p[^ݒ
void set_param(CGAparam &param, CGAunit &unit, char *pname, char *pval)
{
#define SETPARAM_INT(var) { if (!strcmp(pname, #var)) param.m_##var = atoi(pval); }
#define SETPARAM_DBL(var) { if (!strcmp(pname, #var)) param.m_##var = atof(pval); }
#define SETPARAM_BOL(var) { if (!strcmp(pname, #var)) param.m_##var = atoi(pval) ? true : false; }
#define SETPARAM_STAT(var) { if (!strncmp(pname, #var, strlen(#var)) && strlen(pname) == strlen(#var) + 2) param.m_##var[param.str2idx(pname + strlen(#var) + 1)] = atoi(pval); }
	// LN^[
	SETPARAM_INT(chara);
	SETPARAM_INT(style);
	SETPARAM_INT(level);
	// ڕWXe[^X
	SETPARAM_STAT(stat_min);
	SETPARAM_STAT(stat_max);
	SETPARAM_STAT(stat_wei);
	SETPARAM_DBL(stat_wei_hp);
	SETPARAM_DBL(stat_wei_mp);
	// Xe[^X
	SETPARAM_BOL(init_use);
	SETPARAM_INT(init_level);
	SETPARAM_INT(init_point);
	SETPARAM_STAT(init_stat);
	// XienEXe[^X
	if (!strcmp(pname, "ignored_xien")) param.m_ignored_xien |= 1 << param.str2idx(pval);
	if (!strcmp(pname, "ignored_stat")) param.m_ignored_stat |= 1 << param.str2idx(pval);
	// ʃIvV
	SETPARAM_BOL(opt_output_fix);
	SETPARAM_INT(opt_random_rate);
	SETPARAM_DBL(opt_penalty_xien);
	SETPARAM_DBL(opt_penalty_stat);
	// GAIvV
	SETPARAM_INT(ga_try_time);
	SETPARAM_INT(ga_generation);
	SETPARAM_INT(ga_unit_elite);
	SETPARAM_INT(ga_unit_twice);
	SETPARAM_INT(ga_unit_unif);
	SETPARAM_INT(ga_unit_swap);
	SETPARAM_INT(ga_rank_param);
	SETPARAM_DBL(ga_mutate_rate);
//	// ĕzOǂݍ
	if (!strcmp(pname, "import_gasim"))     import_log(param, unit, pval, 0);
	if (!strcmp(pname, "import_oresim"))    import_log(param, unit, pval, 1);
	if (!strcmp(pname, "import_saibunpai")) import_log(param, unit, pval, 2);
	if (!strcmp(pname, "import_talessimu")) import_log(param, unit, pval, 3);
}

void import_log(CGAparam &param, CGAunit &unit, char *logname, int mode)
{
	FILE *fp = fopen(logname, "r");
	if (!fp) {
		printf("warning : Ot@CJ܂ : %s\n", logname);
		return;
	}
	char log[65536];
	fread(log, 1, 65536, fp);
	fclose(fp);

	GAimex *gaimex = NULL;
	switch (mode) {
	case 0: gaimex = (GAimex *)new GAimex_gasim;     break;
	case 1: gaimex = (GAimex *)new GAimex_oresim;    break;
	case 2: gaimex = (GAimex *)new GAimex_saibunpai; break;
	case 3: gaimex = (GAimex *)new GAimex_talessimu; break;
	}
	gaimex->import_log(param, unit, log);
	delete gaimex;

	is_imported = true;
}

void debug_print(CGAparam &param)
{
#define DBGPRINT_INT(var) printf("  %-16s %d\n", #var, param.m_##var);
#define DBGPRINT_DBL(var) printf("  %-16s %f\n", #var, param.m_##var);
#define DBGPRINT_BOL(var) printf("  %-16s %s\n", #var, param.m_##var ?  "on" : "off");

	printf("----------------------------------------------------------------------\n");
	printf("LN^[ :\n");
	DBGPRINT_INT(chara);
	DBGPRINT_INT(style);
	DBGPRINT_INT(level);
	printf("ڕWXe[^X :\n");
	for (int i=0; i<NUM_STAT; i++) {
		printf("  %-4s  %3d %3d %3d\n",
			param.stat_name(i),
			param.m_stat_min[i],
			param.m_stat_max[i],
			param.m_stat_wei[i]
		);
	}
	DBGPRINT_DBL(stat_wei_hp);
	DBGPRINT_DBL(stat_wei_mp);
	printf("Xe[^X :\n");
	DBGPRINT_BOL(init_use);
	DBGPRINT_INT(init_level);
	DBGPRINT_INT(init_point);
	for (int i=0; i<NUM_STAT; i++) {
		printf("%s%3d%s", "  ", param.m_init_stat[i], i == NUM_STAT - 1 ? "\n" : "");
	}
	printf("XienEXe[^X :\n");
	for (int i=0; i<NUM_XIEN; i++) if (param.m_ignored_xien & (1 << i)) printf("  ignored %s\n", param.xien_name(i));
	for (int i=0; i<NUM_STAT; i++) if (param.m_ignored_stat & (1 << i)) printf("  ignored %s\n", param.stat_name(i));
	printf("ʃIvV :\n");
	DBGPRINT_BOL(opt_output_fix);
	DBGPRINT_INT(opt_random_rate);
	DBGPRINT_DBL(opt_penalty_xien);
	DBGPRINT_DBL(opt_penalty_stat);
	printf("GAIvV :\n");
	DBGPRINT_INT(ga_try_time);
	DBGPRINT_INT(ga_generation);
	DBGPRINT_INT(ga_unit_elite);
	DBGPRINT_INT(ga_unit_twice);
	DBGPRINT_INT(ga_unit_unif);
	DBGPRINT_INT(ga_unit_swap);
	DBGPRINT_INT(ga_rank_param);
	DBGPRINT_DBL(ga_mutate_rate);
}
